from powerml import PowerML
import re
from collections import Counter
import random
from tqdm import tqdm
from transformers import GPT2Tokenizer


class PowerMLLearnTopics:
    def __init__(self, config={}, num_subsamples=10, sample_size=40):
        self.model = PowerML(config)
        self.messages = []
        self.num_subsamples = num_subsamples
        self.sample_size = sample_size
        self.memo_top_topics = {}
        self.memo_fltered_top_topics = {}
        self.tokenizer = GPT2Tokenizer.from_pretrained("gpt2")

    def add_data(self, documents):
        self.messages.extend(documents)
        random.shuffle(self.messages)

    def get_topics(self):
        # If get_topics has been called on the same messages array
        # Then use the previous results
        hash_docs = hash(frozenset(self.messages))
        if hash_docs in self.memo_top_topics:
            return self.memo_top_topics[hash_docs]

        num_samples = len(self.messages)
        if self.num_subsamples * self.sample_size > num_samples:
            self.num_subsamples = num_samples // self.sample_size
            self.num_subsamples += 1 if len(
                self.messages) % self.sample_size else 0

        topic_type = 'one-word system components'
        prompt = f'What are the {topic_type} that these messages discuss?\n\n1.'
        num_samples = len(self.messages)
        sample_index = 0
        topic_counter = Counter()
        for i in tqdm(range(self.num_subsamples)):
            if sample_index > num_samples:
                break
            examples, sample_index = self.__subsample(
                self.messages, sample_index, self.sample_size)
            prompt = self.__add_examples(examples, prompt)
            output = self.model.predict(
                prompt, max_tokens=256, temperature=0.7)
            topics = self.__parse_output(output)
            topic_counter.update(topics)
            sorted_topics = self.__sort_topics(topic_counter)
        top_topics = self.__get_top_topics(sorted_topics)
        self.memo_top_topics[hash_docs] = top_topics
        return top_topics

    def get_filtered_topics(self):
        hash_docs = hash(frozenset(self.messages))
        if hash_docs in self.memo_fltered_top_topics:
            return self.memo_fltered_top_topics[hash_docs]
        top_topics = self.get_topics()
        not_topics = self.__filter_top_topics(top_topics)
        filtered_top_topics = list(set(top_topics) - set(not_topics))
        self.memo_fltered_top_topics[hash_docs] = filtered_top_topics
        return filtered_top_topics

    def __count_tokens(self, string):
        return len(self.tokenizer(string)['input_ids'])

    def __add_examples(self, examples, prompt, max_total_tokens=4000, max_output_tokens=256):
        for example in examples:
            new_prompt = f'"{example}"\n\n' + prompt
            if self.__count_tokens(new_prompt) < max_total_tokens - max_output_tokens:
                prompt = new_prompt
            else:
                break
        return prompt

    def __parse_output(self, output):
        list_pattern = re.compile(r"\d+\.\s")
        # include enumerated list prompt
        items = list_pattern.sub("", f'1. {output}')
        parsed = []
        for i in items.split('\n'):
            ii = i.split(',')
            stripped = [iii.strip().replace('.', '') for iii in ii if iii]
            parsed.extend(stripped)
        return parsed

    def __subsample(self, data, start_index, sample_size):
        # Assume data is pre-shuffled
        end_index = start_index + sample_size
        return data[start_index:end_index], end_index

    def __sort_topics(self, topic_counter):
        return sorted(topic_counter.items(), key=lambda x: x[1], reverse=True)

    def __get_top_topics(self, sorted_topics):
        # Return and save all that have more than count 1
        return [elt for elt, count in sorted_topics if count > 1]

    def __filter_top_topics(self, top_topics):
        prompt = 'Which of the above are not technical system-level topics?'
        prompt = f'{top_topics}\n\n{prompt}\n\n1.'
        output = self.model.predict(prompt, max_tokens=500, temperature=0.7)
        output = self.__parse_output(output)
        return output




from powerml import PowerML
import logging

logger = logging.getLogger(__name__)


class PowerMLTopicModel:
    def __init__(self, topics=None, config={}, model_name=None):
        self.model = PowerML(config)
        self.model_name = model_name
        self.topics = topics

        if (topics is None) and (model_name is None):
            raise ValueError("Must specify topics or model_name")

    def fit(self, examples):
        if self.topics is None:
            raise ValueError("Must specify topics to call fit")

        prefix = "List of topics we should consider extracting from messages:"
        example_string = prefix + "\n" + "\n".join(self.topics)

        for example in examples:
            new_string = "\n\nMessage: " + example["example"]
            new_string += "\nExtract the relevant topics from the above message:"
            for label in example["labels"]:
                new_string += "\n-" + label

            if len(example_string) + len(new_string) > 3072:
                break

            example_string += new_string

        suffix = """
Message: {{input}}
Extract the relevant topics from the above message:
-"""
        example_string += suffix

        new_model = self.model.fit(
            [example_string], model="text-davinci-003")

        logger.debug("Got new model: " + new_model)
        # Note: model_name is also stored in the PowerML class.
        # This step is no longer strictly necessary unless multiple models are
        # being used or you wish to explicitly switch between models.
        self.model_name = new_model["model_name"]

    def predict(self, prompt):
        if self.model_name is None:
            result = self.model.predict(prompt)
        else:
            result = self.model.predict(prompt, model=self.model_name)

        return post_process(result)


def post_process(topics):
    return [topic.strip() for topic in topics.split("-")]



import io
import requests
import jsonlines
from powerml.utils.run_ai import run_ai
from powerml.utils.config import get_config
import logging

logger = logging.getLogger(__name__)


class PowerML:
    def __init__(self, config={}):
        self.config = get_config(config)
        self.current_model = "text-davinci-003"

    def predict(self,
                prompt: str,
                model: str = "",
                stop: str = "",
                max_tokens: int = 128,
                temperature: int = 0,
                ) -> str:
        if model == "":
            model = self.current_model
        logger.debug("Predict using model: " + model)
        # if the model is one of our models, then hit our api
        return run_ai(prompt,
                      max_tokens=max_tokens,
                      api="powerml",
                      model=model,
                      stop=stop,
                      temperature=temperature,
                      key=self.config["powerml.key"],
                      )

    def fit(self,
            data: list[str],
            model: str = ""):
        if model == "":
            model = self.current_model
        logger.debug("Fit using model: " + model)
        # Upload filtered data to train api
        headers = {
            "Authorization": "Bearer " + self.config["powerml.key"],
            "Content-Type": "application/json", }
        response = requests.post(
            headers=headers,
            url="https://api.staging.powerml.co/v1/train",
            json={
                "dataset": self.__make_dataset_string(data),
                "model": model
            })
        model_details = response.json()['model']
        self.current_model = model_details["model_name"]
        return model_details

    def __make_dataset_string(self, training_data):
        string = io.StringIO()
        with jsonlines.Writer(string) as writer:
            for item in training_data:
                writer.write({"prompt": item})
        return string.getvalue()



import functools
import csv
import os


def powerml_data(func):
    @functools.wraps(func)
    def wrapper():
        print("\nGenerating_data: " + func.__name__)
        result = func()
        return result
    wrapper.is_generator_function = True
    return wrapper



import functools


def powerml_unit_test_suite(func):
    @functools.wraps(func)
    def wrapper():
        print("\nTest Name: " + func.__name__)
        metric_result, num_tests = func()
        print("Result: " + str(round(metric_result*100)) +
              '%' + " over "+str(num_tests)+" test cases.")
    return wrapper


def powerml_unit_test(func):
    @functools.wraps(func)
    def wrapper():
        print("\nTest Name: " + func.__name__)
        test_result = func()
        if test_result:
            print("Test " + func.__name__ + " passed!")
        else:
            print("Test " + func.__name__ + " failed!")
    return wrapper


def powerml_metric(func):
    @functools.wraps(func)
    def wrapper():
        print("\nMetric Name: " + func.__name__)
        metric_result = func()
        # Expect information about the metric to be non-boolean.
        print("Result: " + str(round(metric_result*100)) + '%' + " passed!")
    return wrapper



from powerml.utils.utils import run_and_postprocess


def fuzzy_coherence_check(text, return_reasons=False):
    '''
    Uses LLM to check whether text has incoherences.

    text: Input text to check.
    '''
    prompt_append = f'\n\nIn the above text, is there anything incoherent? Answer yes or no. If yes, list the incoherences in bullet points.'
    prompt = text + prompt_append

    bool_check, reasons = run_and_postprocess(
        prompt,
        return_extra=return_reasons,
        uses_bool_bullet=True,
        max_tokens=200,
    )

    # Make sure to flip the bool b/c it was used for checking incoherence
    if return_reasons:
        return not bool_check, reasons
    else:
        return not bool_check


if __name__ == "__main__":
    texts = [
        'I like carrots and llamas.',
        '''CX: I'd like to order.
AI: Sorry we are out of that.
CX: Sorry we are really really out of that.
Order: 1x Vanilla Syrup''',
    ]

    for text in texts:
        bool_check, reasons = fuzzy_coherence_check(text, return_reasons=True)

        print(f'Text: {text}')
        print(f'Is it coherent? {"Yes" if bool_check else "No"}')
        print(f'If no, the reasons are {reasons}')
        print('=============================')



from powerml.utils.utils import run_and_postprocess
import numpy as np


def fuzzy_assert_equals(text1, text2):
    '''
    Uses LLM to check whether text1 and text2 are about the same thing.
    '''
    prompt_append = f'Compare the two texts above. Are they the same thing? Answer Yes or No.'
    prompt = f'Text 1: {text1}\n\nText 2: {text2}\n\n{prompt_append}'

    is_equal = run_and_postprocess(prompt,
                                   uses_bool_none=False,
                                   )
    return is_equal


def fuzzy_assert_equals_with_probs(text1, text2):
    '''
    Uses LLM to check whether text1 and text2 are about the same thing.
    '''
    prompt_append = f'Compare the two texts above. Are they the same thing? Answer Yes or No.'
    prompt = f'Text 1: {text1}\n\nText 2: {text2}\n\n{prompt_append}'

    is_equal, logprobs = run_and_postprocess(prompt,
                                             return_top_tokens=True,
                                             uses_bool_none=False,
                                             )
    probs = []
    for token, logprob in logprobs:
        # Only keep yes and no tokens, and compute their probs from logprobs
        if token.lower() not in ['yes', 'no']:
            continue
        prob = np.exp(logprob)
        probs.append((token, prob))

    return is_equal, probs


if __name__ == "__main__":
    texts1 = [
        'Diet coke',
        'extra habenero sauce',
        'cinnamon roll shake',
    ]

    texts2 = [
        'Del coke zero',
        'extra hashbrown sauce',
        'cinnamon red sauce',
    ]

    for text1, text2 in zip(texts1, texts2):
        bool_check, probs = fuzzy_assert_equals(
            text1, text2, return_probs=True)

        print(f'Text 1: {text1}')
        print(f'Text 2: {text2}')
        print(
            f'Are these about the same thing? {"Yes" if bool_check else "No"}')
        print(f'Probs are {probs}')
        print('=============================')



import config
import os


def get_config(dictionary={}):
    global_config = setup_config(dictionary)
    return global_config


def setup_config(dictionary):
    return config.ConfigurationSet(
        config.config_from_dict(dictionary),
        home_yaml_config(),
    )


def home_yaml_config():
    home = os.path.expanduser("~")
    home_config_path = os.path.join(home, ".powerml/configure.yaml")
    if os.path.exists(home_config_path):
        yaml_config = config.config_from_yaml(
            home_config_path, read_from_file=True)
    else:
        yaml_config = config.config_from_dict({})
    return yaml_config



from powerml.utils.run_ai import run_ai
import re


def more_like_this(example, n=1):
    '''
    Uses LLM to get more text like the example.
    '''
    prompt_append = f'Write %d more example(s) written about the same topic but by a different person from the above. List them below.\n1. ' % (
        n)
    prompt = f'Example: {example}\n\n{prompt_append}'
    text = run_ai(prompt,
                  max_tokens=128,
                  )
    results = re.split("\d+\.", text)
    results = [r.strip() for r in results]
    return results


def sql_from_columns(columns=[], n=1):
    '''
    Uses LLM to get more sql queries like the example.
    '''
    column_list_string = "[" + ", ".join(columns) + "]"
    prompt = f'\"\"\"\nTable customers, columns = {column_list_string}\nWrite a complex SQL query for the customers table. Terminate with a semicolon\n\"\"\"\nselect'
    # print(prompt)
    text = run_ai(prompt,
                  max_tokens=128,
                  stop=["\"", ";", "#"],
                  model="code-davinci-002",
                  api="openai",
                  temperature=1,
                  )
    results = re.split("\d+\.", text)
    results = [r.strip() for r in results]
    return results


if __name__ == "__main__":
    texts1 = [
        'AI: Hi, what can we make fresh for you today?\nCX: Chicken Cheddar Quesadilla.',
        'extra habenero sauce',
        'cinnamon roll shake',
    ]

    for text1 in texts1:
        text = more_like_this(
            text1, n=1)

        print(f'Example: {text1}')
        print(f'Similar generated output: {text}')
